from django.db import connection
from django.db.backends import utils as db_backends_utils
from django.db.migrations.state import ModelState
from django.db.models import options
from django.core.exceptions import FieldDoesNotExist

from .contenttypes import ct
from .helpers import is_fake_model


SRC_ATTNAME = 'gm2m_src'
TGT_ATTNAME = 'gm2m_tgt'
CT_ATTNAME = 'gm2m_ct'
FK_ATTNAME = 'gm2m_pk'

THROUGH_FIELDS = (SRC_ATTNAME, TGT_ATTNAME, CT_ATTNAME, FK_ATTNAME)


class Options(options.Options):

    def get_field(self, field_name):
        """
        This override is needed in migrations.RenameModel.database_forwards
        See issue #37
        """
        try:
            return super(Options, self).get_field(field_name)
        except FieldDoesNotExist as e:
            if field_name == self._forward_fields_map[SRC_ATTNAME]. \
                                  related_model._meta.model_name:
                return super(Options, self).get_field(SRC_ATTNAME)
            raise e


def create_gm2m_intermediary_model(field, klass):
    """
    Creates a generic M2M model for the GM2M field 'field' on model 'klass'
    """

    from django.db import models

    managed = klass._meta.managed
    name = '%s_%s' % (klass._meta.object_name, field.name)

    model_name = klass._meta.model_name

    db_table = db_backends_utils.truncate_name(
                   '%s_%s' % (klass._meta.db_table, field.name),
                   connection.ops.max_name_length())

    meta_kwargs = {
        'db_table': db_table,
        'managed': managed,
        'auto_created': klass,
        'app_label': klass._meta.app_label,
        'db_tablespace': klass._meta.db_tablespace,
        'unique_together': (SRC_ATTNAME, CT_ATTNAME, FK_ATTNAME),
        'verbose_name': '%s-generic relationship' % model_name,
        'verbose_name_plural': '%s-generic relationships' % model_name,
        'apps': field.model._meta.apps,
    }

    meta = type('Meta', (object,), meta_kwargs)

    fk_maxlength = 16  # default value
    if field.pk_maxlength is not False:
        fk_maxlength = field.pk_maxlength

    model = type(str(name), (models.Model,), {
        'Meta': meta,
        '__module__': klass.__module__,
        SRC_ATTNAME: models.ForeignKey(
            klass, auto_created=True,
            on_delete=field.remote_field.on_delete_src,
            db_constraint=field.remote_field.db_constraint
        ),
        CT_ATTNAME: models.ForeignKey(
            ct.ContentType,
            on_delete=models.CASCADE,
            db_constraint=field.remote_field.db_constraint
        ),
        FK_ATTNAME: models.CharField(max_length=fk_maxlength),
        TGT_ATTNAME: ct.GenericForeignKey(
            ct_field=CT_ATTNAME,
            fk_field=FK_ATTNAME,
            for_concrete_model=field.remote_field.for_concrete_model,
        ),
    })

    if is_fake_model(klass):
        # if we are building a fake model for migrations purposes, create a
        # ModelState from the model and render it (see issues #3 and #5)
        model = ModelState.from_model(model).render(klass._meta.apps)

    # changing the options' class to override get_field for RenameModel
    # migrations' database_forwards
    model._meta.__class__ = Options

    return model



from django.db.models.fields.related import \
    ForeignObjectRel, ForeignObject, ManyToManyRel, lazy_related_operation
from django.core.exceptions import FieldDoesNotExist
from django.db.models.signals import pre_delete
from django.db.utils import DEFAULT_DB_ALIAS
from django.db.models import Q
from django.apps import apps
from django.core import checks
from django.utils.functional import cached_property
from django.db.models.query_utils import PathInfo

from .contenttypes import ct, get_content_type
from .models import create_gm2m_intermediary_model, THROUGH_FIELDS
from .managers import create_gm2m_related_manager
from .descriptors import RelatedGM2MDescriptor, SourceGM2MDescriptor
from .deletion import *
from .signals import deleting
from .helpers import GM2MModel, is_fake_model


# default relation attributes
# we set that here as they are used to retrieve the GM2MRel attribute from
# a GM2MUnitRel (see GM2MUnitRel.__getattribute__ below)
# they are also used in GM2MField.deconstruct

# attributes whose values can be overriden by kwargs
REL_ATTRS = {
    'related_name': None,
    'related_query_name': None,
    'through': None,
    'through_fields': None,
    'db_constraint': True,
    'for_concrete_model': True,
    'on_delete': CASCADE,
}

# attributes whose values cannot be changed
REL_ATTRS_FIXED = {
    'multiple': True,
    'symmetrical': False,
    'parent_link': False,
    'limit_choices_to': {}
}

REL_ATTRS_NAMES = list(REL_ATTRS.keys()) + list(REL_ATTRS_FIXED.keys()) + [
    'on_delete_tgt', 'on_delete_src'
]


class GM2MRelation(ForeignObject):
    """
    A reverse relation for a GM2MField.
    Each related model has a GM2MRelation towards the source model
    """

    # copies GM2MField flags (as self.field will always be a GM2MField)
    many_to_many = True
    many_to_one = False
    one_to_many = False
    one_to_one = False

    concrete = False
    related_accessor_class = RelatedGM2MDescriptor

    hidden = False

    def __init__(self, model, field, rel, **kwargs):
        self.field = field
        self.model = model

        kwargs.update({
            'rel': rel,
            'name': rel.get_accessor_name() + '_relation',
            'blank': True,
            'editable': False,
            'serialize': False,
            'on_delete': rel.on_delete
        })

        super(GM2MRelation, self).__init__(model, from_fields=[field.name],
                                           to_fields=[], **kwargs)

    def contribute_to_class(self, cls, name, virtual_only=False):
        pass

    def get_accessor_name(self):
        return self.remote_field.get_accessor_name()

    def bulk_related_objects(self, objs, using=DEFAULT_DB_ALIAS):
        """
        Return all objects related to objs
        The returned result will be passed to Collector.collect, so one should
        not use the deletion functions as such
        """

        through = self.field.remote_field.through
        base_mngr = through._base_manager.db_manager(using)

        on_delete = self.remote_field.on_delete

        if on_delete is not DO_NOTHING:
            # collect related objects
            field_names = through._meta._field_names
            q = Q()
            for obj in objs:
                # Convert each obj to (content_type, primary_key)
                q = q | Q(**{
                    field_names['tgt_ct']: get_content_type(obj),
                    field_names['tgt_fk']: obj.pk
                })
            qs = base_mngr.filter(q)

            if on_delete in (DO_NOTHING_SIGNAL, CASCADE_SIGNAL,
                             CASCADE_SIGNAL_VETO):
                results = deleting.send(sender=self.field,
                                        del_objs=objs, rel_objs=qs)
            else:
                results = []

            if on_delete in (CASCADE, CASCADE_SIGNAL) \
            or on_delete is CASCADE_SIGNAL_VETO \
            and not any(r[1] for r in results):
                # if CASCADE must be called or if no receiver returned a veto
                # we return the qs for deletion
                # note that it is an homogeneous queryset (as Collector.collect
                # which is called afterwards only works with homogeneous
                # collections)
                return qs

        # do not delete anything by default
        empty_qs = base_mngr.none()
        empty_qs.query.set_empty()
        return empty_qs


class GM2MUnitRel(ForeignObjectRel):

    dummy_pre_delete = lambda s, **kwargs: None

    def __init__(self, field, model, auto, on_delete=None):
        super(GM2MUnitRel, self).__init__(field, model, on_delete=on_delete)
        self.multiple = True
        # warning: do NOT use self.auto_created as it's used by Django !!
        self.auto = auto

    def check(self, **kwargs):
        errors = []
        errors.extend(self._check_referencing_to_swapped_model())
        errors.extend(self._check_clashes())
        return errors

    def _check_referencing_to_swapped_model(self):
        if self.model not in apps.get_models() \
        and not isinstance(self.model, str) \
        and self.model._meta.swapped:
            model = '%s.%s' % (
                self.model._meta.app_label,
                self.model._meta.object_name
            )
            return [checks.Error(
                ("Field defines a relation with the model '%s', "
                 "which has been swapped out.") % model,
                hint="Update the relation to point at 'settings.%s'."
                     % self.remote_field.model._meta.swappable,
                obj=self,
                id='gm2m.E201',
            )]
        return []

    def _check_clashes(self):
        """ Check accessor and reverse query name clashes. """

        from django.db.models.base import ModelBase

        errors = []
        opts = self.field.model._meta

        # `self.model` may be a string instead of a model. Skip if model name
        # is not resolved.
        if not isinstance(self.model, ModelBase):
            return []

        # If the field doesn't install backward relation on the target
        # model (so `is_hidden` returns True), then there are no clashes to
        # check and we can skip these fields.
        if self.is_hidden():
            return []

        try:
            self.related
        except AttributeError:
            return []

        # Consider that we are checking field `Model.gm2m` and the
        # models are:
        #
        #     class Target(models.Model):
        #         model = models.IntegerField()
        #         model_set = models.IntegerField()
        #
        #     class Model(models.Model):
        #         foreign = models.ForeignKey(Target, on_delete=models.CASCADE)
        #         gm2m = GM2MField(Target)

        rel_opts = self.model._meta
        # rel_opts.object_name == "Target"
        rel_name = self.related.get_accessor_name()  # i. e. "model_set"
        rel_query_name = self.field.related_query_name()  # i. e. "model"
        field_name = "%s.%s" % (opts.object_name,
                                self.field.name)  # i. e. "Model.gm2m"

        # Check clashes between accessor or reverse query name of `field`
        # and any other field name -- i. e. accessor for Model.gm2m is
        # model_set and it clashes with Target.model_set.
        potential_clashes = rel_opts.fields + rel_opts.many_to_many
        for clash_field in potential_clashes:
            clash_name = "%s.%s" % (rel_opts.object_name,
                clash_field.name)  # i. e. "Target.model_set"
            if clash_field.name == rel_name:
                errors.append(
                    checks.Error(
                        "Reverse accessor for '%s' clashes with field "
                        "name '%s'." % (field_name, clash_name),
                        hint="Rename field '%s', or add/change a "
                             "related_name argument to the definition "
                             "for field '%s'." % (clash_name, field_name),
                        obj=self,
                        id='gm2m.E202',
                    )
                )

            if clash_field.name == rel_query_name:
                errors.append(
                    checks.Error(
                        "Reverse query name for '%s' clashes with field "
                        "name '%s'." % (field_name, clash_name),
                        hint="Rename field '%s', or add/change a "
                             "related_name argument to the definition "
                             "for field '%s'." % (clash_name, field_name),
                        obj=self,
                        id='gm2m.E203',
                    )
                )

        # Check clashes between accessors/reverse query names of `field`
        # and any other field accessor -- i. e. Model.gm2m accessor clashes
        # with Model.foreign accessor.
        potential_clashes = [r for r in rel_opts.related_objects
                             if r.field is not self.field]
        for clash_field in potential_clashes:
            # "Model.gm2m"
            clash_name = "%s.%s" % (
                clash_field.model._meta.object_name,
                clash_field.field.name)
            if clash_field.get_accessor_name() == rel_name:
                errors.append(
                    checks.Error(
                        "Reverse accessor for '%s' clashes with reverse "
                        "accessor for '%s'." % (field_name, clash_name),
                        hint=("Add or change a related_name argument "
                              "to the definition for '%s' or '%s'.")
                             % (field_name, clash_name),
                        obj=self,
                        id='gm2m.E204',
                    )
                )

            if clash_field.get_accessor_name() == rel_query_name:
                errors.append(
                    checks.Error(
                        "Reverse query name for '%s' clashes with reverse "
                        "query name for '%s'." % (field_name, clash_name),
                        hint=("Add or change a related_name argument "
                              "to the definition for '%s' or '%s'.")
                             % (field_name, clash_name),
                        obj=self,
                        id='gm2m.E205',
                    )
                )

        return errors

    def __getattribute__(self, name):
        """
        General attributes are those from the GM2MRel object
        """
        sup = super(GM2MUnitRel, self).__getattribute__
        if name in REL_ATTRS_NAMES:
            if name == 'on_delete':
                # try and get locally defined on_delete, otherwise use the
                # remote_field's on_delete_tgt
                on_delete = sup(name)
                if on_delete is None:
                    name += '_tgt'
                else:
                    return on_delete
            return getattr(sup('field').remote_field, name)
        else:
            return sup(name)

    def contribute_to_class(self):
        if isinstance(self.model, str) or self.model._meta.pk is None:
            def resolve_related_class(cls, model, rel):
                rel.model = model
                rel.do_related_class()
            lazy_related_operation(resolve_related_class, self.field.model,
                                   self.model, rel=self)
        else:
            self.do_related_class()

    def do_related_class(self):
        # check that the relation does not already exist
        all_rels = self.field.remote_field.rels
        if self.model in [r.model for r in all_rels if r != self]:
            # if it does, it needs to be removed from the list, and no further
            # action should be taken
            all_rels.remove(self)
            return

        self.related = GM2MRelation(self.field.model, self.field, self)
        if not self.field.model._meta.abstract:
            self.contribute_to_related_class()

    def contribute_to_related_class(self):
        """
        Appends accessors to related classes
        """

        # this enables cascade deletion for any relation (even hidden ones)
        self.model._meta.add_field(self.related, private=True)

        if self.on_delete in handlers_with_signal:
            # if a signal should be sent on deletion, we connect a dummy
            # receiver to pre_delete so that the model is not
            # 'fast_delete'-able
            # (see django.db.models.deletion.Collector.can_fast_delete)
            pre_delete.connect(self.dummy_pre_delete, sender=self.model)

        # Internal M2Ms (i.e., those with a related name ending with '+')
        # and swapped models don't get a related descriptor.
        if not self.is_hidden() and not self.field.model._meta.swapped:
            setattr(self.model, self.related_name
                        or (self.field.model._meta.model_name + '_set'),
                    RelatedGM2MDescriptor(self.related, self))

    @cached_property
    def related_manager_cls(self):
        # the related manager class getter is implemented here rather than in
        # the descriptor as we may need to access it even for hidden relations
        return create_gm2m_related_manager(
            superclass=self.field.model._default_manager.__class__,
            field=self.related.field,
            model=self.field.model,
            through=self.through,
            query_field_name=self.through._meta.model_name,
            field_names=self.through._meta._field_names,
            prefetch_cache_name=self.related.field.related_query_name()
        )

    @property
    def swappable_setting(self):
        """
        Gets the setting that this is powered from for swapping, or None
        if it's not swapped in
        """
        # can only be called by Django 1.7+, the apps module will be available

        # Work out string form of "to"
        if isinstance(self.model, str):
            to_string = self.model
        else:
            to_string = "%s.%s" % (
                self.model._meta.app_label,
                self.model._meta.object_name,
            )
        # See if anything swapped/swappable matches
        for model in apps.get_models(include_swapped=True):
            if model._meta.swapped == to_string \
            or model._meta.swappable \
            and ("%s.%s" % (model._meta.app_label,
                            model._meta.object_name)) == to_string:
                return model._meta.swappable
        return None

    def _get_path_info(self, filtered_relation, reverse):
        pathinfos = []

        opts = self.through._meta

        # this is the src <> through part of the relation, we'll use
        # path info retrieval functions on this
        fk_field = opts.get_field(opts._field_names['src'])

        if reverse:
            pathinfos.extend(fk_field.get_reverse_path_info())
            # through > to part of the relation is generated manually
            opts = self.model._meta
            pathinfos.append(PathInfo(self.through._meta, opts, (opts.pk,),
                                      self, True, False, filtered_relation))
        else:
            # to > through part of the relation is generated manually
            opts = self.through._meta
            pathinfos.append(PathInfo(self.model._meta, opts, (opts.pk,),
                                      self, False, False, filtered_relation))
            pathinfos.extend(fk_field.get_path_info())
        return pathinfos

    def get_path_info(self, filtered_relation=None):
        return self._get_path_info(filtered_relation, reverse=False)

    def get_reverse_path_info(self, filtered_relation=None):
        return self._get_path_info(filtered_relation, reverse=True)

    def get_joining_columns(self):
        opts = self.through._meta
        return [(
            self.model._meta.pk.column,
            opts.get_field(opts._field_names['tgt_fk']).column
        )]

    def get_extra_restriction(self, where_class, alias, remote_alias):
        opts = self.through._meta
        field = opts.get_field(opts._field_names['tgt_ct'])

        ct_pk = ct.ContentType.objects.get_for_model(self.model,
                    for_concrete_model=self.for_concrete_model).pk
        lookup = field.get_lookup('exact')(field.get_col(alias), ct_pk)

        cond = where_class()
        cond.add(lookup, 'AND')
        return cond

    def get_related_field(self):
        """
        Returns the field in the to object to which this relationship is tied
        (this is always the primary key on the target model).
        """
        return self.model._meta.pk


class GM2MRel(ManyToManyRel):

    model = GM2MModel

    name = 'gm2mrel'
    is_relation = False
    hidden = False

    def __init__(self, field, related_models, **params):

        self.field = field
        self._init_attrs = {}

        for name, default in REL_ATTRS.items():
            self.set_init(name, params.pop(name, default))
        for name, value in REL_ATTRS_FIXED.items():
            self.set_init(name, value)

        for on_del_param in ('src', 'tgt'):
            on_del_param = 'on_delete_' + on_del_param
            self.set_init(
                on_del_param, params.pop(on_del_param, self.on_delete)
            )

        if self.through and not self.db_constraint:
            raise ValueError('django-gm2m: Can\'t supply a through model '
                             'with db_constraint=False')

        self.rels = []
        for model in related_models:
            self.add_relation(model, contribute_to_class=False)

    def set_init(self, name, value):
        super(GM2MRel, self).__setattr__(name, value)

    def __setattr__(self, name, value):
        # setting a keyword attribute afterwards should not have any influence
        # on the deconstruction. Hence we
        if name in REL_ATTRS_NAMES and name not in self._init_attrs:
            self._init_attrs[name] = getattr(self, name)
        self.set_init(name, value)

    def add_relation(self, model, on_delete=None, auto=False,
                     contribute_to_class=True):
        try:
            assert not model._meta.abstract, \
            "%s cannot define a relation with abstract class %s" \
            % (self.field.__class__.__name__, model._meta.object_name)
        except AttributeError:
            # model._meta doesn't exist, so it must be a string
            assert isinstance(model, str), \
            '%s(%r) is invalid. First parameter to GM2MField must ' \
            'be either a model or a model name' \
            % (self.field.__class__.__name__, model)

        rel = GM2MUnitRel(self.field, model, auto=auto, on_delete=on_delete)
        self.rels.append(rel)
        if contribute_to_class:
            rel.contribute_to_class()
        return rel

    def check(self, **kwargs):
        errors = []
        for rel in self.rels:
            errors.extend(rel.check(**kwargs))
        errors.extend(self._check_relationship_model(**kwargs))
        return errors

    def _check_relationship_model(self, from_model=None, **kwargs):
        if hasattr(self.through, '_meta'):
            qualified_model_name = "%s.%s" % (
                self.through._meta.app_label, self.through.__name__)
        else:
            qualified_model_name = self.through

        errors = []

        if self.through not in apps.get_models(include_auto_created=True):
            # The relationship model is not installed.
            errors.append(
                checks.Error(
                    ("Field specifies a many-to-many relation through model "
                     "'%s', which has not been installed.") %
                    qualified_model_name,
                    hint=None,
                    obj=self,
                    id='gm2m.E101',
                )
            )

        else:

            assert from_model is not None, \
                "GM2MField with intermediate " \
                "tables cannot be checked if you don't pass the model " \
                "where the field is attached to."

            # Set some useful local variables
            from_model_name = from_model._meta.object_name

            # Count foreign keys in intermediate model
            seen_from = sum(
                from_model == getattr(field.remote_field, 'model', None)
                for field in self.through._meta.fields
            )

            if seen_from == 0:
                errors.append(
                    checks.Error(
                        ("The model is used as an intermediate model by '%s', "
                         "but it does not have a foreign key to '%s' or a "
                         "generic foreign key.") % (self, from_model_name),
                        hint=None,
                        obj=self.through,
                        id='gm2m.E102',
                    )
                )
            elif seen_from > 1 and not self.through_fields:
                errors.append(
                    checks.Warning(
                        "The model is used as an intermediate model by "
                        "'%s', but it has more than one foreign key "
                        "from '%s', which is ambiguous. You must specify "
                        "which foreign key Django should use via the "
                        "through_fields keyword argument."
                        % (self, from_model_name),
                        hint=None,
                        obj=self,
                        id='gm2m.E103',
                    )
                )

            seen_to = sum(isinstance(field, ct.GenericForeignKey)
                for field in self.through._meta.private_fields)

            if seen_to == 0:
                errors.append(
                    checks.Error(
                        "The model is used as an intermediate model by "
                        "'%s', but it does not have a a generic foreign key."
                        % from_model_name,
                        hint=None,
                        obj=self.through,
                        id='gm2m.E104',
                    )
                )
            elif seen_to > 1 and not self.through_fields:
                errors.append(
                    checks.Warning(
                        "The model is used as an intermediate model by "
                        "'%s', but it has more than one generic foreign "
                        "key, which is ambiguous. You must specify "
                        "which generic foreign key Django should use via "
                        "the through_fields keyword argument." % self,
                        hint=None,
                        obj=self,
                        id='gm2m.E105',
                    )
                )

        # Validate `through_fields`
        if self.through_fields is not None:
            # Validate that we're given an iterable of at least two items
            # and that none of them is "falsy"
            if not (len(self.through_fields) >= 2 and
                    self.through_fields[0] and self.through_fields[1]):
                errors.append(
                    checks.Error(
                        ("Field specifies 'through_fields' but does not "
                         "provide the names of the two link fields that "
                         "should be used for the relation through model "
                         "'%s'.") % qualified_model_name,
                        hint=("Make sure you specify 'through_fields' as "
                              "through_fields=('field1', 'field2')"),
                        obj=self,
                        id='gm2m.E106',
                    )
                )

            # Validate the given through fields -- they should be actual
            # fields on the through model, and also be foreign keys to the
            # expected models
            else:
                assert from_model is not None, \
                    "GM2MField with intermediate " \
                    "tables cannot be checked if you don't pass the model " \
                    "where the field is attached to."

                src_field_name = self.through_fields[0]
                through = self.through

                possible_field_names = []
                for f in through._meta.fields:
                    if hasattr(f, 'remote_field') \
                    and getattr(f.remote_field, 'model', None) == from_model:
                        possible_field_names.append(f.name)
                if possible_field_names:
                    hint = ("Did you mean one of the following foreign keys "
                            "to '%s': %s?") % (from_model._meta.object_name,
                                               ', '.join(possible_field_names))
                else:
                    hint = None

                try:
                    field = through._meta.get_field(src_field_name)
                except FieldDoesNotExist:
                    errors.append(
                        checks.Error(
                            "The intermediary model '%s' has no field '%s'."
                            % (qualified_model_name, src_field_name),
                            hint=hint,
                            obj=self,
                            id='gm2m.E107',
                        )
                    )
                else:
                    if not (getattr(field, 'remote_field', None) and
                            getattr(field.remote_field, 'model', None) == from_model):
                        errors.append(
                            checks.Error(
                                "'%s.%s' is not a foreign key to '%s'." % (
                                    through._meta.object_name, src_field_name,
                                    from_model._meta.object_name),
                                hint=hint,
                                obj=self,
                                id='gm2m.E108',
                            )
                        )

                target_field_name = self.through_fields[1]

                possible_field_names = []
                for f in through._meta.private_fields:
                    if isinstance(f, ct.GenericForeignKey):
                        possible_field_names.append(f.name)
                if possible_field_names:
                    hint = "Did you mean one of the following generic " \
                           "foreign keys: %s?" \
                           % ', '.join(possible_field_names)
                else:
                    hint = None

                field = None
                for f in through._meta.private_fields:
                    if f.name == target_field_name:
                        field = f
                        break
                else:
                    errors.append(
                        checks.Error(
                            "The intermediary model '%s' has no generic "
                            "foreign key named '%s'."
                            % (qualified_model_name, src_field_name),
                            hint=hint,
                            obj=self,
                            id='gm2m.E109',
                        )
                    )

                if field:
                    if not isinstance(field, ct.GenericForeignKey):
                        errors.append(
                            checks.Error(
                                "'%s.%s' is not a generic foreign key."
                                % (through._meta.object_name, src_field_name),
                                hint=hint,
                                obj=self,
                                id='gm2m.E110',
                            )
                        )
        return errors

    def contribute_to_class(self, cls, virtual_only=False):

        # Connect the descriptor for this field
        setattr(cls, self.field.attname,
                SourceGM2MDescriptor(self.field))

        if cls._meta.abstract or cls._meta.swapped:
            # do not do anything for abstract or swapped model classes
            return

        if not self.through:
            self.set_init('through',
                          create_gm2m_intermediary_model(self.field, cls))
            # we set through_fields to the default intermediary model's
            # THROUGH_FIELDS as it carries fields assignments for
            # ModelState instances
            self.set_init('through_fields', THROUGH_FIELDS)

        # set related name
        if not self.field.model._meta.abstract and self.related_name:
            self.set_init('related_name', self.related_name % {
                'class': self.field.model.__name__.lower(),
                'app_label': self.field.model._meta.app_label.lower()
            })

        def calc_field_names(rel):
            # Extract field names from through model and stores them in
            # rel.through_field (so that they are sent on deconstruct and
            # passed to ModelState instances)

            tf_dict = {}

            if is_fake_model(rel.through):
                # we populate the through field dict using rel.through_fields
                # that was either provided or computed beforehand with the
                # actual model
                for f, k in zip(rel.through_fields,
                                ('src', 'tgt', 'tgt_ct', 'tgt_fk')):
                    tf_dict[k] = f
                rel.through._meta._field_names = tf_dict
                return

            if rel.through_fields:
                tf_dict['src'], tf_dict['tgt'] = \
                    rel.through_fields[:2]
                for gfk in rel.through._meta.private_fields:
                    if gfk.name == tf_dict['tgt']:
                        break
                else:
                    raise FieldDoesNotExist(
                        'Generic foreign key "%s" does not exist in through '
                        'model "%s"' % (tf_dict['tgt'],
                                        rel.through._meta.model_name)
                    )
                tf_dict['tgt_ct'] = gfk.ct_field
                tf_dict['tgt_fk'] = gfk.fk_field
            else:
                for f in rel.through._meta.fields:
                    try:
                        remote_field = f.remote_field
                    except AttributeError:
                        continue
                    if remote_field and (remote_field.model == rel.field.model
                        or remote_field.model == '%s.%s' % (
                            rel.field.model._meta.app_label,
                            rel.field.model._meta.object_name)):
                        tf_dict['src'] = f.name
                        break
                for f in rel.through._meta.private_fields:
                    if isinstance(f, ct.GenericForeignKey):
                        tf_dict['tgt'] = f.name
                        tf_dict['tgt_ct'] = f.ct_field
                        tf_dict['tgt_fk'] = f.fk_field
                        break

            if not set(tf_dict.keys()).issuperset(('src', 'tgt')):
                raise ValueError('Bad through model for GM2M relationship.')

            rel.through._meta._field_names = tf_dict

            # save the result in rel.through_fields so that it appears
            # in the deconstruction. Without that there would be no way for
            # a ModelState constructed from a migration to know which fields
            # have which function, as all virtual fields are stripped
            tf = []
            for f in ('src', 'tgt', 'tgt_ct', 'tgt_fk'):
                tf.append(tf_dict[f])
            rel.set_init('through_fields', tf)

        # resolve through model if it's provided as a string
        if isinstance(self.through, str):
            def resolve_through_model(c, model, r):
                r.set_init('through', model)
                calc_field_names(r)
            lazy_related_operation(resolve_through_model, cls, self.through,
                                   r=self)
        else:
            calc_field_names(self)

        self.related_model = cls

        for rel in self.rels:
            # we need to make sure the GM2MUnitRel's field instance is the
            # right one. Indeed, if cls is derived from an abstract model
            # where the GM2MField is defined, rel.field is the field linked
            # to the abstract model
            rel.field = self.field
            rel.contribute_to_class()

    @cached_property
    def related_manager_cls(self):
        field_names = self.through._meta._field_names
        return create_gm2m_related_manager(
            superclass=None,
            field=self.field,
            model=self.through,
            through=self.through,
            query_field_name=field_names['src'],
            field_names=field_names,
            prefetch_cache_name=self.field.name
        )

    def get_accessor_name(self):
        return self.related_name \
            or (self.field.model._meta.model_name + '_set')



from django.db.migrations.state import StateApps
from django.utils.functional import cached_property


class _CTClasses(object):

    @cached_property
    def ContentType(self):
        from django.contrib.contenttypes.models import ContentType
        return ContentType

    @cached_property
    def GenericForeignKey(self):
        from django.contrib.contenttypes.fields import GenericForeignKey
        return GenericForeignKey


ct = _CTClasses()


def get_content_type(obj):

    try:
        # obj is a model instance, retrieve database
        db = obj._state.db
        klass = obj.__class__
    except AttributeError:
        # obj is a model class
        db = None
        klass = obj

    ct_mngr = ct.ContentType.objects.db_manager(db)
    if isinstance(klass._meta.apps, StateApps):
        # if obj is an instance of a fake model for migrations purposes, use
        # ContentType's ModelState rather than ContentType itself (issue #14)
        # this should not raise LookupError as at this stage contenttypes must
        # be loaded
        ct_mngr.model = obj._meta.apps.get_model('contenttypes', 'ContentType')
        # we erase the app cache to make sure a modelstate is returned when
        # calling get_for_model on the manager
        try:
            del ct_mngr._cache[db][(klass._meta.app_label, klass._meta.model_name)]
        except KeyError:
            pass
    else:
        ct_mngr.model = ct.ContentType

    return ct_mngr.get_for_model(obj)



from django.core.serializers import json
from . import python


class Serializer(python.Serializer, json.Serializer):
    """
    As python.Serializer does not override anything from json.Serializer,
    nothing wrong with using the MRO
    """
    pass

Deserializer = json.Deserializer



import subprocess

__version_info__ = (1, 1, 1, 'final', 0)


def get_version(version=__version_info__):

    dev_st = {'alpha': 'a', 'beta': 'b', 'rc': 'c', 'final': ''}

    assert len(version) == 5
    assert version[3] in dev_st.keys()

    n = 2 + (version[2] != 0)
    version_str = '.'.join([str(v) for v in version[:n]])

    if version[3] == 'final':
        return version_str

    if version[3:] == ('alpha', 0):
        return '%s.dev0+%s' % (version_str, get_git_chgset())
    else:
        return ''.join((version_str, dev_st[version[3]], str(version[4])))


def get_git_chgset():
    try:
        return subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD'],
                                       universal_newlines=True).strip()[:-1]
    except:
        return '?'


__version__ = get_version()



import warnings

from django.db.models.fields import Field
from django.db import connection
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _
from django.core import checks
from django.db.backends import utils as db_backends_utils

from .relations import GM2MRel, REL_ATTRS, REL_ATTRS_NAMES


class GM2MField(Field):
    """
    Provides a generic relation to several generic objects through a
    generic model storing content-type/object-id information

    Reverse relations can be established with models provided as arguments
    """

    # Field flags
    auto_created = False
    concrete = False
    editable = False
    hidden = False

    is_relation = True
    many_to_many = True
    many_to_one = False
    one_to_many = False
    one_to_one = False

    related_model = ''

    description = _('Generic many-to-many relationship')

    def __init__(self, *related_models, **params):

        super(GM2MField, self).__init__(
            verbose_name=params.pop('verbose_name', None),
            name=params.pop('name', None),
            help_text=params.pop('help_text', u''),
            error_messages=params.pop('error_messages', None),
            rel=GM2MRel(self, related_models, **params),
            blank=params.pop('blank', False),
            # setting null to True only prevent makemigrations from asking for
            # a default value
            null=True
        )

        self.db_table = params.pop('db_table', None)
        self.pk_maxlength = params.pop('pk_maxlength', False)
        if self.remote_field.through is not None:
            assert self.db_table is None and self.pk_maxlength is False, \
                'django-gm2m: Cannot specify a db_table nor a pk_maxlength ' \
                'if ''an intermediary model is used.'

    def check(self, **kwargs):
        errors = super(GM2MField, self).check(**kwargs)
        errors.extend(self._check_unique(**kwargs))
        errors.extend(self.remote_field.check(**kwargs))
        return errors

    def _check_unique(self, **kwargs):
        if self.unique:
            return [
                checks.Error(
                    'GM2MFields cannot be unique.',
                    hint=None,
                    obj=self,
                    id='gm2m.E001',
                )
            ]
        return []

    def deconstruct(self):
        name, path, args, kwargs = super(GM2MField, self).deconstruct()

        kwargs.pop('null', None)

        # generate related models list (cannot get it from rel, as it can
        # be changed by add_relation)
        for rel in self.remote_field.rels:
            if getattr(rel, '_added', False):
                continue

            if isinstance(rel.model, str):
                args.append(rel.model)
            else:
                # see if the related model is a swappable model
                swappable_setting = rel.swappable_setting
                if swappable_setting is not None:
                    setting_name = getattr(rel.model, 'setting_name', None)
                    if setting_name != swappable_setting:
                        raise ValueError(
                            'Cannot deconstruct a GM2MField pointing to a '
                            'model that is swapped in place of more than one '
                            'model (%s and %s)'
                            % (setting_name, swappable_setting))

                    from django.db.migrations.writer import SettingsReference
                    model = SettingsReference(rel.model, swappable_setting)
                else:
                    model = '%s.%s' % (rel.model._meta.app_label,
                                    rel.model._meta.object_name)
                args.append(model)

        # handle parameters
        if self.db_table:
            kwargs['db_table'] = self.db_table
        if self.pk_maxlength is not False:
            kwargs['pk_maxlength'] = self.pk_maxlength

        through = self.remote_field.through
        if through:
            if isinstance(through, str):
                kwargs['through'] = through
            elif not through._meta.auto_created:
                kwargs['through'] = '%s.%s' % (through._meta.app_label,
                                               through._meta.object_name)

        # rel options
        for k in REL_ATTRS_NAMES:
            if k == 'through':
                # through has been dealt with just above
                continue

            # retrieve default value
            try:
                default = REL_ATTRS[k]
            except KeyError:
                if k.startswith('on_delete_'):
                    default = kwargs.get('on_delete', REL_ATTRS['on_delete'])
                else:
                    # this is a fixed attribute, we don't need to care about it
                    continue

            # retrieve actual initial value, possibly from _init_attr dict
            try:
                value = self.remote_field._init_attrs[k]
            except KeyError:
                value = getattr(self.remote_field, k)

            if value != default:
                if k == 'related_name':
                    value = force_text(value)
                kwargs[k] = value

        return name, path, args, kwargs

    def add_relation(self, model, on_delete=None, auto=False):
        rel = self.remote_field.add_relation(model, on_delete=on_delete,
                                             auto=auto)
        rel._added = True

    def get_related_models(self, include_auto=False):
        models = []
        for unitrel in self.remote_field.rels:
            if not unitrel.auto or include_auto:
                models.append(unitrel.model)
        return models

    def db_type(self, connection):
        """
        A GM2M field will not have a column as it defines a relation between
        tables
        """
        return None

    def get_internal_type(self):
        """
        A GM2M field behaves like a ManyToManyField
        """
        # For Django 1.7
        return 'ManyToManyField'

    def m2m_db_table(self):
        # self.db_table will be None if
        if self.remote_field.through is not None:
            return self.remote_field.through._meta.db_table
        elif self.db_table:
            return self.db_table
        else:
            return db_backends_utils.truncate_name(
                '%s_%s' % (self.model._meta.db_table, self.name),
                connection.ops.max_name_length())

    def contribute_to_class(self, cls, name, virtual_only=False):
        """
        Appends accessor to a class and create intermediary model

        :param cls: the class to contribute to
        :param name: the name of the accessor
        """
        self.set_attributes_from_name(name)
        self.model = cls

        opts = cls._meta
        if virtual_only:
            opts.add_virtual_field(self)
        else:
            opts.add_field(self)
            # we need to clear the options cache here. It would automatically
            # be done by add_field above if GM2MField derived from
            # ManyToManyField, but for various reasons it is not the case
            opts._expire_cache()

        self.model = cls
        self.opts = cls._meta

        # Set up related classes if relations are defined
        self.remote_field.contribute_to_class(cls)

    def get_attname_column(self):
        """
        A GM2M field will not have a column as it defines a relation between
        tables
        """
        attname = self.get_attname()
        return attname, None

    def is_hidden(self):
        """
        Should the related object be hidden?
        """
        return self.remote_field.related_name \
            and self.remote_field.related_name[-1] == '+'

    def related_query_name(self):
        return self.remote_field.related_query_name \
            or self.remote_field.related_name \
            or self.model._meta.model_name

    def formfield(self, form_class=None, choices_form_class=None, **kwargs):
        warnings.warn(
            'gm2m.GM2MField values cannot be represented in a form field yet. '
            '"%s" will be ignored' % self.name, FutureWarning
        )
        return None



from django.core.serializers import pyyaml
from . import python


class Serializer(python.Serializer, pyyaml.Serializer):
    """
    As python.Serializer does not override anything from pyyaml.Serializer,
    nothing wrong with using the MRO
    """
    pass

Deserializer = pyyaml.Deserializer



"""
Generic many-to-many relations descriptors
"""


class GM2MDescriptor(object):

    def __init__(self, field):
        self.field = field

    def __get__(self, instance, instance_type=None):
        if instance is None:
            return self
        return self.related_manager_cls(instance)

    def __set__(self, instance, value):
        if not self.through._meta.auto_created:
            opts = self.through._meta
            raise AttributeError(
                'Cannot set values on a GM2MField which specifies an '
                'intermediary model. Use %s.%s\'s Manager instead.'
                % (opts.app_label, opts.object_name))
        manager = self.__get__(instance)
        manager.set(value)


class RelatedGM2MDescriptor(GM2MDescriptor):
    """
    Provides a generic many-to-many descriptor for the related models to make
    the source manager available from a target model class
    """

    def __init__(self, related, remote_field):
        super(RelatedGM2MDescriptor, self).__init__(related)
        self.remote_field = remote_field

    @property
    def through(self):
        return self.remote_field.through

    @property
    def related_manager_cls(self):
        return self.remote_field.related_manager_cls


class SourceGM2MDescriptor(GM2MDescriptor):
    """
    Provides a generic many-to-many descriptor for the source model to make the
    related manager available from the source model class, and to access field
    management methods
    """

    def add_relation(self, model, on_delete=None, auto=False):
        return self.field.add_relation(model, on_delete=on_delete, auto=auto)

    def get_related_models(self, include_auto=False):
        return self.field.get_related_models(include_auto)

    @property
    def through(self):
        return self.field.remote_field.through

    @property
    def related_manager_cls(self):
        return self.field.remote_field.related_manager_cls

    def __set__(self, instance, value):
        # clear() can change expected output of 'value' queryset,
        # we force evaluation of queryset before clear; django ticket #19816
        value = tuple(value)
        super(SourceGM2MDescriptor, self).__set__(instance, value)



"""
Yes, monkey-patching Django is not the greatest thing to do, but in that
case there is no other solution to make it alter *both* fields needed by the
GFK (content type + primary key)
"""

import django
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.backends.sqlite3.schema import DatabaseSchemaEditor
from django.db.migrations.autodetector import MigrationAutodetector
from django.db.migrations.operations.models import RenameModel


# ALL BACKENDS EXCEPT SQLITE

_alter_many_to_many_0 = BaseDatabaseSchemaEditor._alter_many_to_many


def _alter_many_to_many(self, model, old_field, new_field, strict):
    from .fields import GM2MField
    if isinstance(old_field, GM2MField) \
    or isinstance(new_field, GM2MField):
        # Rename the through table
        if old_field.remote_field.through._meta.db_table != \
           new_field.remote_field.through._meta.db_table:
            self.alter_db_table(old_field.remote_field.through,
                                old_field.remote_field.through._meta.db_table,
                                new_field.remote_field.through._meta.db_table)
        # Repoint the GFK to the other side
        # we need to alter both fields of the GFK
        old_names = old_field.remote_field.through._meta._field_names
        new_names = new_field.remote_field.through._meta._field_names
        getoldfield = old_field.remote_field.through._meta.get_field
        getnewfield = new_field.remote_field.through._meta.get_field
        self.alter_field(
            new_field.remote_field.through,
            getoldfield(old_names['tgt_fk']),
            getnewfield(new_names['tgt_fk']),
        )
        self.alter_field(
            new_field.remote_field.through,
            getoldfield(old_names['tgt_ct']),
            getnewfield(new_names['tgt_ct']),
        )
        # now we alter the fk
        self.alter_field(
            new_field.remote_field.through,
            getoldfield(old_names['src']),
            getnewfield(new_names['src']),
        )
    else:
        return _alter_many_to_many_0(self, model, old_field, new_field,
                                     strict)
BaseDatabaseSchemaEditor._alter_many_to_many = _alter_many_to_many


# SQLITE BACKEND, SPECIFIC IMPLEMENTATION

_alter_many_to_many_sqlite0 = DatabaseSchemaEditor._alter_many_to_many


def _alter_many_to_many(self, model, old_field, new_field, strict):
    from .fields import GM2MField
    if isinstance(old_field, GM2MField) \
    or isinstance(new_field, GM2MField):
        # Repoint the GFK to the other side
        # we need to alter both fields of the GFK
        old_names = old_field.remote_field.through._meta._field_names
        new_names = new_field.remote_field.through._meta._field_names
        getoldfield = old_field.remote_field.through._meta.get_field
        getnewfield = new_field.remote_field.through._meta.get_field
        
        if old_field.remote_field.through._meta.db_table == \
        new_field.remote_field.through._meta.db_table:
            # The field name didn't change, but some options did;
            # we have to propagate this altering.
            # We need the field that points to the target model,
            # so we can tell alter_field to change it -
            # this is m2m_reverse_field_name() (as opposed to
            # m2m_field_name, which points to our model)
            for f in ['tgt_fk', 'tgt_ct']:
                self._remake_table(
                    old_field.remote_field.through,
                    alter_field=(
                        getoldfield(old_names[f]),
                        getnewfield(new_names[f]),
                    )
                )
            return

        # Make a new through table
        self.create_model(new_field.remote_field.through)
        # Copy the data across
        self.execute("INSERT INTO %s (%s) SELECT %s FROM %s" % (
            self.quote_name(new_field.remote_field.through._meta.db_table),
            ', '.join([
                "id",
                getnewfield(old_names['src']).column,
                getnewfield(old_names['tgt_fk']).column,
                getnewfield(old_names['tgt_ct']).column,
            ]),
            ', '.join([
                "id",
                getoldfield(old_names['src']).column,
                getoldfield(old_names['tgt_fk']).column,
                getoldfield(old_names['tgt_ct']).column,
            ]),
            self.quote_name(old_field.remote_field.through._meta.db_table),
        ))
        # Delete the old through table
        self.delete_model(old_field.remote_field.through)
    else:
        return _alter_many_to_many_sqlite0(self, model, old_field,
                                           new_field, strict)


DatabaseSchemaEditor._alter_many_to_many = _alter_many_to_many


def only_relation_agnostic_fields(self, fields):
    """
    We only change the way the 'model' key is deleted from the dict
    (as GM2MField.deconstruct does not return a 'model' kwarg)
    """
    fields_def = []

    fields_items = None
    if django.VERSION >= (3, 1):
        fields_items = sorted(fields.items())
    else:
        fields_items = sorted(fields)

    for __, field in fields_items:
        deconstruction = self.deep_deconstruct(field)
        if field.remote_field and field.remote_field.model:
            deconstruction[2].pop('model', None)
        fields_def.append(deconstruction)
    return fields_def

MigrationAutodetector.only_relation_agnostic_fields = \
    only_relation_agnostic_fields


database_forwards0 = RenameModel.database_forwards

def database_forwards(self, app_label, schema_editor, from_state, to_state):
    """
    Filter out auto created related objects from Options.related_objects
    """

    old_model = from_state.apps.get_model(app_label, self.old_name)
    related_objects_bck = old_model._meta.related_objects
    old_model._meta.related_objects = [o for o in related_objects_bck
                                       if not o.auto_created]

    database_forwards0(self, app_label, schema_editor, from_state, to_state)

    old_model._meta.related_objects = related_objects_bck

RenameModel.database_forwards = database_forwards



import django
from django.db import router
from django.db.models import Q, Manager
from django.db import connections

from .contenttypes import ct, get_content_type
from .query import GM2MTgtQuerySet


class GM2MBaseManager(Manager):

    use_in_migration = True

    def __init__(self, instance):
        super(GM2MBaseManager, self).__init__()
        self.model = self._model  # see create_gm2m_related_manager
        self.instance = instance
        self.pk = instance.pk
        self.core_filters = {}

    def get_queryset(self):
        try:
            return self.instance \
                       ._prefetched_objects_cache[self.prefetch_cache_name]
        except (AttributeError, KeyError):
            db = self._db or router.db_for_read(self.instance.__class__,
                                                instance=self.instance)
            return self._get_queryset(using=db)._next_is_sticky() \
                       .filter(**self.core_filters)

    def _get_queryset(self, using):
        return super(GM2MBaseManager, self).get_queryset().using(using)

    def get_prefetch_queryset(self, instances, queryset=None):
        db = self._db or router.db_for_read(self.model,
                                            instance=instances[0])

        if queryset is None:
            queryset = self._get_queryset(db)

        qs, rel_obj_attr, instance_attr = \
            self._get_prefetch_queryset_params(instances, queryset, db)

        return (qs,
                rel_obj_attr,
                instance_attr,
                False,
                self.prefetch_cache_name,
                False)

    def _get_extra_queryset(self, queryset, q, extra_fields, db):
        join_table = self.through._meta.db_table
        connection = connections[db]
        qn = connection.ops.quote_name
        extra = dict(select=dict(
            ('_prefetch_related_val_%s' % f.attname,
             '%s.%s' % (qn(join_table), qn(f.column)))
            for f in extra_fields))
        return queryset.using(db)._next_is_sticky().filter(q).extra(**extra)

    def _check_through_model(self, method_name):
        # If the GM2M relation has an intermediary model,
        # the add and remove methods are not available.
        if not self.through._meta.auto_created:
            opts = self.through._meta
            raise AttributeError(
                'Cannot use %s() on a ManyToManyField which specifies an '
                'intermediary model. Use %s.%s\'s Manager instead.'
                % (method_name, opts.app_label, opts.object_name))

    def _do_add(self, db, through_objs):
        """
        Performs items addition
        """
        # Add the new entries in the db table
        self.through._default_manager.using(db).bulk_create(through_objs)

    def add(self, *objs):
        """
        Adds objects to the GM2M field
        :param *objs: object instances to add
        """
        #

        self._check_through_model('add')

        if not objs:
            return

        db = router.db_for_write(self.through, instance=self.instance)
        self._do_add(db, self._to_add(objs, db))

    add.alters_data = True

    def _do_remove(self, db, q):
        """
        Perfoms items removal from a Q object
        """
        self.through._default_manager.using(db).filter(q).delete()

    def remove(self, *objs):
        """
        Removes objects from the GM2M field
        """
        # *objs - objects to remove

        self._check_through_model('remove')

        if not objs:
            return

        db = router.db_for_write(self.through, instance=self.instance)
        self._do_remove(db, self._to_remove(objs))
    remove.alters_data = True

    def _do_clear(self, db, filter=None):
        self.through._default_manager.using(db).filter(**(filter or {})) \
            .delete()

    def set(self, objs, **kwargs):
        """
        Replaces the set of related objects by the items in the objs iterable
        """

        self._check_through_model('set')

        objs = tuple(objs)

        clear = kwargs.pop('clear', False)
        db = router.db_for_write(self.through, instance=self.instance)

        if clear:
            # clears all and re-adds
            self._do_clear(db)
            self._do_add(db, *objs)
        else:
            # just removes the necessary items and adds the missing ones
            to_add, to_remove = self._to_change(objs, db)
            self._do_remove(db, to_remove)
            self._do_add(db, to_add)
    set.alters_data = True

    def clear(self):
        db = router.db_for_write(self.through, instance=self.instance)
        self._do_clear(db, self._to_clear())

    clear.alters_data = True


class GM2MBaseSrcManager(Manager):
    
    def __init__(self, instance):
        # the manager's model is the source model
        super(GM2MBaseSrcManager, self).__init__(instance)
        self.core_filters['%s__%s' % (self.query_field_name,
                                      self.field_names['tgt_ct'])] = \
            get_content_type(self.instance)
        self.core_filters['%s__%s' % (self.query_field_name,
                                      self.field_names['tgt_fk'])] = \
            self.instance.pk

    def _get_prefetch_queryset_params(self, instances, queryset, db):

        # we're looking for generic target instances, which should be
        # converted to (content_type, primary_key) tuples

        q = Q()
        for obj in instances:
            q = q | Q(**{
                '%s__%s' % (self.query_field_name,
                            self.field_names['tgt_ct']):get_content_type(obj),
                '%s__%s' % (self.query_field_name,
                            self.field_names['tgt_fk']): obj.pk
            })

        # Annotating the query in order to retrieve the primary model
        # content type and id in the same query
        # content type must be the 1st element, see rel_obj_attr below
        extra_fields = (
            self.through._meta.get_field(self.field_names['tgt_ct']),
            self.through._meta.get_field(self.field_names['tgt_fk'])
        )

        qs = self._get_extra_queryset(queryset, q, extra_fields, db)

        # primary model retrieval function
        def rel_obj_attr(relobj):
            t = []
            for f in extra_fields:
                try:
                    # t already contains the content type id
                    # we use get_for_id to retrieve the cached content type
                    model = ct.ContentType.objects.get_for_id(t[0]) \
                                                  .model_class()
                except IndexError:
                    # t is empty
                    model = ct.ContentType
                t.append(model._meta.pk.to_python(
                    getattr(relobj, '_prefetch_related_val_%s' % f.attname)
                ))
            return tuple(t)

        # model attribute retrieval function
        instance_attr = lambda inst: \
            (get_content_type(inst).pk, inst.pk)

        return qs, rel_obj_attr, instance_attr

    def _to_add(self, objs, db):
        # we're using the reverse relation to add source model
        # instances
        inst_ct = get_content_type(self.instance)
        vals = self.through._default_manager.using(db) \
                           .values_list(self.field_names['src'],
                                        flat=True) \
                           .filter(**{
                               self.field_names['tgt_ct']: inst_ct,
                               self.field_names['tgt_fk']: self.pk
                           })
        to_add = []
        for obj in objs:
            if obj.pk not in vals:
                to_add.append(self.through(**{
                    '%s_id' % self.field_names['src']:
                        obj.pk,
                    self.field_names['tgt_ct']: inst_ct,
                    self.field_names['tgt_fk']: self.pk
                }))
        return to_add

    def _to_remove(self, objs):
        # we're using the reverse relation to delete source model
        # instances
        inst_ct = get_content_type(self.instance)
        return Q(**{
            '%s_id__in' % self.field_names['src']:
                [obj.pk for obj in objs],
            self.field_names['tgt_ct']: inst_ct,
            self.field_names['tgt_fk']: self.pk
        })

    def _to_change(self, objs, db):
        """
        Returns the sets of items to be added and a Q object for removal
        """
        inst_ct = get_content_type(self.instance)
        
        vals = list(self.through._default_manager.using(db)
                                .values_list(self.field_names['src'], flat=True)
                                .filter(**{
                                    self.field_names['tgt_ct']: inst_ct,
                                    self.field_names['tgt_fk']: self.pk
                                }))
        
        to_add = set()
        to_remove = set()
        for obj in objs:
            try:
                vals.remove(obj.pk)
            except ValueError:
                # obj.pk is not in vals and must be added
                to_add.add(self.through(**{
                    '%s_id' % self.field_names['src']:
                        obj.pk,
                    self.field_names['tgt_ct']: inst_ct,
                    self.field_names['tgt_fk']: self.pk
                }))

        for v in vals:
            to_remove.add(v)
        
        return to_add, Q(pk__in=to_remove)

    def _to_clear(self):
        return {
            self.field_names['tgt_ct']: get_content_type(self.instance),
            self.field_names['tgt_fk']: self.instance.pk
        }


class GM2MBaseTgtManager(Manager):

    def __init__(self, instance):
        # the manager's model is the through model
        super(GM2MBaseTgtManager, self).__init__(instance)

        source_field = self.through._meta.get_field(
                           self.field_names['src'])
        self.source_related_fields = source_field.related_fields
        for __, rh_field in self.source_related_fields:
            key = '%s__%s' % (self.query_field_name, rh_field.name)
            self.core_filters[key] = getattr(self.instance,
                                             rh_field.attname)

    def _get_queryset(self, using):
        return GM2MTgtQuerySet(self.model, using=using)

    def _get_prefetch_queryset_params(self, instances, queryset, db):

        # we're looking for through model instances
        query = {}
        for lh_field, rh_field in self.source_related_fields:
            query['%s__in' % lh_field.name] = \
                set(getattr(obj, rh_field.attname)
                    for obj in instances)
        q = Q(**query)

        # Annotating the query in order to retrieve the primary model
        # id in the same query
        fk = self.through._meta.get_field(self.field_names['src'])
        extra_fields = fk.local_related_fields

        qs = self._get_extra_queryset(queryset, q, extra_fields, db)
        # marking the queryset so that the original queryset should
        # be returned when evaluated the first time
        qs._related_prefetching = True

        # primary model retrieval function
        def rel_obj_attr(relobj):
            t = []
            for f in extra_fields:
                v = getattr(relobj,
                            '_prefetch_related_val_%s' % f.attname)
                try:
                    v = v.pop()
                except AttributeError:  # v is not a list
                    pass
                t.append(f.related_model._meta.pk.to_python(v))
            return tuple(t)

        # model attribute retrieval function
        select_fields = fk.foreign_related_fields
        instance_attr = lambda inst: tuple([getattr(inst, f.attname)
                            for f in select_fields])

        return qs, rel_obj_attr, instance_attr

    def _to_add(self, objs, db):
        models = []
        objs_set = set()
        for obj in objs:
            # extract content type and primary key for each object
            objs_set.add((get_content_type(obj),
                          obj.pk))
            m = obj.__class__
            if m not in models:
                # call field.add_relation for each model
                models.append(m)
                self.field.add_relation(m, auto=True)

        vals = self.through._default_manager.using(db) \
                   .filter(**{self.field_names['src']: self.pk}) \
                   .values_list(self.field_names['tgt_ct'],
                                self.field_names['tgt_fk'])

        to_add = []
        for ct, pk in objs_set.difference(vals):
            to_add.append(self.through(**{
                '%s_id' % self.field_names['src']: self.pk,
                self.field_names['tgt_ct']: ct,
                self.field_names['tgt_fk']: pk
            }))

        return to_add

    def _to_remove(self, objs):
        q = Q()
        for obj in objs:
            # Convert the obj to (content_type, primary_key)
            q = q | Q(**{
                self.field_names['tgt_ct']: get_content_type(obj).pk,
                self.field_names['tgt_fk']: obj.pk
            })
        return q & Q(**{
            '%s_id' % self.field_names['src']: self.pk
        })

    def _to_clear(self):
        return {
            '%s_id' % self.field_names['src']: self.pk
        }

    def _to_change(self, objs, db):
        """
        Returns the sets of items to be added and a Q object for removal
        """

        to_add = set()

        src_fname = self.field_names['src']
        ct_fname = self.field_names['tgt_ct']
        fk_fname = self.field_names['tgt_fk']

        vals = list(self.through._default_manager.using(db)
                        .filter(**{self.field_names['src']: self.pk})
                        .values_list(ct_fname, fk_fname))

        known_cts = set(v[0] for v in vals)

        for obj in objs:
            ct = get_content_type(obj)
            val = (ct, obj.pk)
            try:
                vals.remove(val)
            except ValueError:
                # val is not in vals
                # extract content type and primary key for each object
                to_add.add((ct, obj.pk))

                if ct.pk not in known_cts:
                    # call field.add_relation for each unknown model
                    self.field.add_relation(obj.__class__, auto=True)
                    known_cts.add(ct.pk)

        rem_q = Q()
        for val in vals:
            # Convert the obj to (content_type, primary_key)
            rem_q = rem_q | Q(**{
                ct_fname: val[0],
                fk_fname: val[1]
            })

        return [
            self.through(**{
                '%s_id' % src_fname: self.pk,
                ct_fname: t[0],
                fk_fname: t[1]
            }) for t in to_add
        ], \
        rem_q & Q(**{
            '%s_id' % src_fname: self.pk
        })


def create_gm2m_related_manager(superclass=None, **kwargs):
    """
    Dynamically create a manager class that only concerns an instance (source
    or target)
    """

    bases = [GM2MBaseManager]

    if superclass is None:
        # no superclass provided, the manager is a generic target model manager
        bases.insert(0, GM2MBaseTgtManager)
    else:
        # superclass provided, the manager is a source model manager and also
        # derives from superclass
        bases.insert(0, GM2MBaseSrcManager)
        bases.append(superclass)

    # Django's Manager constructor sets model to None, we store it under the
    # class's attribute '_model' and it is retrieved in __init__
    kwargs['_model'] = kwargs.pop('model')
    return type(Manager)('GM2MManager', tuple(bases), kwargs)



from django.apps import AppConfig
from django.core import serializers


class GM2MConfig(AppConfig):
    name = 'gm2m'
    verbose_name = 'Django generic many-to-many field'

    def ready(self):

        from . import signals
        from . import monkeypatch

        serializers.BUILTIN_SERIALIZERS = {
            "xml": "gm2m.serializers.xml_serializer",
            "python": "gm2m.serializers.python",
            "json": "gm2m.serializers.json",
            "yaml": "gm2m.serializers.pyyaml",
        }



from django.core.serializers import xml_serializer
from django.utils.encoding import smart_text

from ..fields import GM2MField
from ..contenttypes import ct, get_content_type


class Serializer(xml_serializer.Serializer):
    """
    xml_serializer.Serializer uses a different handle_m2m_field than
    python.Serializer, we therefore need to redefine it here
    """

    def handle_m2m_field(self, obj, field):

        if not isinstance(field, GM2MField):
            # use normal serialization from superclass
            super(Serializer, self).handle_m2m_field(obj, field)

        if field.remote_field.through._meta.auto_created:
            self._start_relational_field(field)
            if self.use_natural_foreign_keys:
                def handle_gm2m(value):
                    try:
                        natural = value.natural_key()
                        use_natural_key = True
                    except AttributeError:
                        natural = smart_text(value._get_pk_val())
                        use_natural_key = False

                    # Iterable natural keys are rolled out as subelements
                    if use_natural_key:
                        attrs = {'pk': smart_text(value._get_pk_val())}
                    else:
                        attrs = {}

                    self.xml.startElement("object", attrs)

                    # add content type information
                    app, model = get_content_type(value).natural_key()
                    self.xml.addQuickElement('contenttype', attrs={
                        'app': app,
                        'model': model
                    })

                    if use_natural_key:
                        for key_value in natural:
                            self.xml.startElement("natural", {})
                            self.xml.characters(smart_text(key_value))
                            self.xml.endElement("natural")

                    self.xml.endElement("object")
            else:
                def handle_gm2m(value):
                    self.xml.startElement('object', {
                        'pk': smart_text(value._get_pk_val())
                    })

                    # add content type information
                    app, model = get_content_type(value).natural_key()
                    self.xml.addQuickElement('contenttype', attrs={
                        'app': app,
                        'model': model
                    })

                    self.xml.endElement('object')

            for relobj in getattr(obj, field.name).iterator():
                handle_gm2m(relobj)

            self.xml.endElement("field")


class Deserializer(xml_serializer.Deserializer):

    def _handle_m2m_field_node(self, node, field):
        """
        Handle a <field> node for a GM2MField
        """

        if not isinstance(field, GM2MField):
            return super(Deserializer, self)._handle_m2m_field_node(node, field)

        objs = []
        for obj_node in node.getElementsByTagName('object'):
            natural = obj_node.getElementsByTagName('natural')

            # extract contenttype
            ct_node = obj_node.getElementsByTagName('contenttype')[0]

            model = ct.ContentType.objects.get_by_natural_key(
                ct_node.getAttribute('app'), ct_node.getAttribute('model')
            ).model_class()
            mngr = model._default_manager.db_manager(self.db)


            if natural:
                # extract natural keys
                key = [xml_serializer.getInnerText(k).strip() for k in natural]
            else:
                # normal value
                key = obj_node.getAttribute('pk')

            if hasattr(model._default_manager, 'get_by_natural_key'):
                if hasattr(key, '__iter__') \
                and not isinstance(key, str):
                    obj = mngr.get_by_natural_key(*key)
                else:
                    obj = mngr.get_by_natural_key(key)
            else:
                obj = mngr.get(pk=key)

            objs.append(obj)

        return objs




from django.db.models.deletion import CASCADE, DO_NOTHING

from .signals import deleting


__all__ = ['CASCADE', 'DO_NOTHING', 'CASCADE_SIGNAL', 'CASCADE_SIGNAL_VETO',
           'DO_NOTHING_SIGNAL', 'handlers_with_signal', 'handlers_do_nothing']


def collector_data_iterator(data):
    for __, instances in data.items():
        for instance in instances:
            yield instance


def CASCADE_SIGNAL(collector, field, sub_objs, using):
    deleting.send(field, del_objs=collector_data_iterator(collector.data),
                  rel_objs=sub_objs)
    CASCADE(collector, field, sub_objs, using)


def CASCADE_SIGNAL_VETO(collector, field, sub_objs, using):
    res = deleting.send(field,
                        del_objs=collector_data_iterator(collector.data),
                        rel_objs=sub_objs)
    if not any(r[1] for r in res):
        # if no receiver returned a truthy result (veto), we can
        # cascade-collect, else we do nothing
        CASCADE(collector, field, sub_objs, using)


def DO_NOTHING_SIGNAL(collector, field, sub_objs, using):
    deleting.send(field, del_objs=collector_data_iterator(collector.data),
                  rel_objs=sub_objs)


handlers_do_nothing = (DO_NOTHING, DO_NOTHING_SIGNAL)
handlers_with_signal = (CASCADE_SIGNAL, CASCADE_SIGNAL_VETO, DO_NOTHING_SIGNAL)



from django.core.serializers import python
from django.utils.encoding import force_text

from ..fields import GM2MField
from ..contenttypes import get_content_type


class Serializer(python.Serializer):

    def handle_m2m_field(self, obj, field):
        if isinstance(field, GM2MField):
            if field.remote_field.through._meta.auto_created:
                if self.use_natural_foreign_keys:
                    def m2m_value(value):
                        try:
                            natural = value.natural_key()
                        except AttributeError:
                            natural = force_text(value._get_pk_val(),
                                             strings_only=True)
                        return (
                            get_content_type(value).natural_key(),
                            natural
                        )
                else:
                    def m2m_value(value):
                        return (
                            get_content_type(value).natural_key(),
                            force_text(value._get_pk_val(), strings_only=True)
                        )
                self._current[field.name] = [m2m_value(related)
                    for related in getattr(obj, field.name).iterator()]
        else:
            # use normal serialization
            super(Serializer, self).handle_m2m_field(obj, field)

Deserializer = python.Deserializer



"""
django-gm2m
Django generic many-to-many field
(c) 2014-2020 Thomas Khyn
MIT License (see LICENSE.txt)
"""

from setuptools import setup, find_packages
import os


# imports __version__ variable
exec(open('gm2m/version.py').read())
dev_status = __version_info__[3]

if dev_status == 'alpha' and not __version_info__[4]:
    dev_status = 'pre'

DEV_STATUS = {'pre': '2 - Pre-Alpha',
              'alpha': '3 - Alpha',
              'beta': '4 - Beta',
              'rc': '4 - Beta',
              'final': '5 - Production/Stable'}

# setup function parameters
setup(
    name='django-gm2m',
    version=__version__,
    description='Django generic many-to-many field',
    long_description=open(os.path.join('README.rst')).read(),
    author='Thomas Khyn',
    author_email='thomas@ksytek.com',
    url='https://github.com/tkhyn/django-gm2m',
    keywords=['django', 'generic', 'many-to-many'],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Development Status :: %s' % DEV_STATUS[dev_status],
        'Framework :: Django',
        'Environment :: Other Environment',
    ],
    packages=find_packages(exclude=('tests',)),
    install_requires=(
        'Django>=2.2',
    ),
    zip_safe=True,
)



from django.db import models
from django.db.models.fields.related import RECURSIVE_RELATIONSHIP_CONSTANT
from django.db.migrations.state import StateApps
from django.utils.functional import cached_property

from .contenttypes import ct


def is_fake_model(model):
    """
    Is ``model`` a 'state' model (generated for migrations)
    """
    return isinstance(model._meta.apps, StateApps)


# Hereafter, the aim is to create a 'dummy' model class that:
#  - enables django to find out that GM2MField depends on contenttypes
#  - provides a specific manager for deserialization

# Indeed, unlike a GFK, GM2MField does not create any FK to ContentType on
# the source model, so we need to let django know about that


class GM2MModelOptions(object):

    def __init__(self):
        self.object_name = 'ContentType'
        self.model_name = 'contenttype'
        self.app_label = 'contenttypes'

        self.concrete_fields = []
        self.pk = None

    def __str__(self):
        return 'gm2m.model'

    @cached_property
    def concrete_model(self):
        return ct.ContentType


class GM2MModelManager(models.Manager):

    def get_by_natural_key(self, ct_key, key):
        """
        Used for deserialization (this is actually a workaround to avoid
        heavy monkey-patching)
        :param ct_key: the content type's natural key
        :param key: the object key (may be a natural key)
        :return:
        """

        model = ct.ContentType.objects.get_by_natural_key(*ct_key).model_class()
        mngr = model._default_manager.db_manager(self.db)

        if hasattr(model._default_manager, 'get_by_natural_key'):
            if hasattr(key, '__iter__') and not isinstance(key, (str, bytes)):
                obj = mngr.get_by_natural_key(*key)
            else:
                obj = mngr.get_by_natural_key(key)
        else:
            obj = mngr.get(pk=key)

        # django's Deserializer only cares about the pk attribute, but we
        # need the actual instance
        gm2mto = GM2MModel()
        gm2mto.pk = obj

        return gm2mto

    def all(self):
        return []


class Dummy(object):
    """
    We can't derive our dummy model class from models.Model explicitly, as if we
    do it triggers all the django machinery through the metaclass. To prevent
    this, we derive from a Dummy class and then overwrite the __bases__ (as we
    are not allowed to replace __bases__ on a class tha derives from `object`
    """


class GM2MModelMetaclass(type):
    # the following methods are necessary for migration
    # RenameModel._get_model_tuple
    # See issue #37
    def __str__(cls):
        return RECURSIVE_RELATIONSHIP_CONSTANT

    def __hash__(cls):
        return hash(RECURSIVE_RELATIONSHIP_CONSTANT)

    def __iter__(cls):
        yield None

    def __eq__(cls, other):
        return other is cls \
               or other == RECURSIVE_RELATIONSHIP_CONSTANT

    def lower(cls):
        return str(cls)


class GM2MModel(Dummy, metaclass=GM2MModelMetaclass):
    """
    We need to define pk as we're using that attribute in the GM2MToManager
    above
    """
    pk = None


GM2MModel.__bases__ = (models.Model,)
GM2MModel._meta = GM2MModelOptions()
GM2MModel._default_manager = GM2MModelManager()



from collections import defaultdict

from django.db.models.query import ModelIterable, QuerySet

from .contenttypes import ct as ct_classes, get_content_type


class GM2MTgtQuerySetIterable(ModelIterable):

    def __iter__(self):
        """
        Override to return the actual objects, not the GM2MObject
        Fetch the actual objects by content types to optimize database access
        """

        qs = self.queryset

        try:
            del qs._related_prefetching
            rel_prefetching = True
        except AttributeError:
            rel_prefetching = False

        ct_attrs = defaultdict(lambda: defaultdict(lambda: []))
        objects = {}
        ordered_ct_attrs = []

        field_names = qs.model._meta._field_names
        fk_field = qs.model._meta.get_field(field_names['tgt_fk'])

        extra_select = list(qs.query.extra_select)

        for vl in qs.values_list(field_names['tgt_ct'],
                                 field_names['tgt_fk'],
                                 *extra_select):
            ct = vl[0]
            pk = fk_field.to_python(vl[1])
            ct_attrs[ct][pk].append(vl[2:])
            ordered_ct_attrs.append((ct, pk))

        for ct, attrs in ct_attrs.items():
            for pk, obj in ct_classes.ContentType.objects.get_for_id(ct).\
                    model_class()._default_manager.in_bulk(attrs.keys()).\
                    items():

                pk = fk_field.to_python(pk)

                # we store the through model id in case we are in the process
                # of fetching related objects
                for i, k in enumerate(extra_select):
                    e_list = []
                    for e in attrs[pk]:
                        e_list.append(e[i])
                    setattr(obj, k, e_list)

                if rel_prefetching:
                    # when prefetching related objects, one must yield one
                    # object per through model instance
                    for __ in attrs[pk]:
                        if qs.ordered:
                            objects[(ct, pk)] = obj
                        else:
                            yield obj
                    continue

                if qs.ordered:
                    objects[(ct, pk)] = obj
                else:
                    yield obj

        if qs.ordered:
            for ct, pk in ordered_ct_attrs:
                yield objects[(ct, pk)]


class GM2MTgtQuerySet(QuerySet):
    """
    A QuerySet for GM2M models which yields actual target generic objects
    instead of GM2M objects when iterated over
    It can also filter the output by model (= content type)
    """

    def __init__(self, model=None, query=None, using=None, hints=None):
        super(GM2MTgtQuerySet, self).__init__(model, query, using, hints)

        if self._iterable_class is ModelIterable:
            self._iterable_class = GM2MTgtQuerySetIterable

    def filter(self, *args, **kwargs):
        model = kwargs.pop('Model', None)
        models = kwargs.pop('Model__in', set())

        if model:
            models.add(model)

        ctypes = []
        for m in models:
            if isinstance(m, str):
                m = self.model._meta.apps.get_model(m)
            ctypes.append(get_content_type(m).pk)

        if ctypes:
            kwargs[self.model._meta._field_names['tgt_ct'] + '__in'] = ctypes

        return super(GM2MTgtQuerySet, self).filter(*args, **kwargs)
